from contextlib import contextmanager, suppress
from contextvars import ContextVar
from functools import wraps
from typing import (
    Any,
    AsyncContextManager,
    Awaitable,
    Callable,
    ContextManager,
    Generator,
    ParamSpec,
    Protocol,
    TypeVar,
)

from jararaca.microservice import AppContext

P = ParamSpec("P")
R = TypeVar("R")


class TracingContextProvider(Protocol):

    def __call__(
        self, trace_name: str, context_attributes: dict[str, str]
    ) -> ContextManager[Any]: ...


class TracingContextProviderFactory(Protocol):

    def root_setup(self, app_context: AppContext) -> AsyncContextManager[None]: ...

    def provide_provider(self, app_context: AppContext) -> TracingContextProvider: ...


tracing_ctx_provider_ctxv = ContextVar[TracingContextProvider]("tracing_ctx_provider")


@contextmanager
def provide_tracing_ctx_provider(
    ctx_provider: TracingContextProvider,
) -> Generator[None, None, None]:

    token = tracing_ctx_provider_ctxv.set(ctx_provider)
    try:
        yield
    finally:
        with suppress(ValueError):
            tracing_ctx_provider_ctxv.reset(token)


def get_tracing_ctx_provider() -> TracingContextProvider | None:
    return tracing_ctx_provider_ctxv.get(None)


def default_trace_mapper(*args: Any, **kwargs: Any) -> dict[str, str]:
    return {
        "args": str(args),
        "kwargs": str(kwargs),
    }


class TracedFunc:

    def __init__(
        self,
        trace_name: str,
        trace_mapper: Callable[..., dict[str, str]] = default_trace_mapper,
    ):
        self.trace_name = trace_name
        self.trace_mapper = trace_mapper

    def __call__(
        self,
        decorated: Callable[P, Awaitable[R]],
    ) -> Callable[P, Awaitable[R]]:

        @wraps(decorated)
        async def wrapper(
            *args: P.args,
            **kwargs: P.kwargs,
        ) -> R:

            if ctx_provider := get_tracing_ctx_provider():
                with ctx_provider(
                    self.trace_name,
                    self.trace_mapper(**kwargs),
                ):
                    return await decorated(*args, **kwargs)

            return await decorated(*args, **kwargs)

        return wrapper



from contextlib import asynccontextmanager
from typing import Any, AsyncContextManager, AsyncGenerator, Protocol

import aio_pika
from pydantic import BaseModel

from jararaca.messagebus.publisher import MessagePublisher, provide_message_publisher
from jararaca.microservice import AppContext, AppInterceptor


class MessageBusConnectionFactory(Protocol):

    def provide_connection(self) -> AsyncContextManager[MessagePublisher]: ...


class MessageBusPublisherInterceptor(AppInterceptor):

    def __init__(
        self,
        connection_factory: MessageBusConnectionFactory,
        connection_name: str = "default",
    ):
        self.connection_factory = connection_factory
        self.connection_name = connection_name

    @asynccontextmanager
    async def intercept(self, app_context: AppContext) -> AsyncGenerator[None, None]:
        async with self.connection_factory.provide_connection() as connection:
            with provide_message_publisher(self.connection_name, connection):
                yield


class AIOPikaMessagePublisher(MessagePublisher):

    def __init__(self, channel: aio_pika.abc.AbstractChannel, exchange_name: str):
        self.channel = channel
        self.exchange_name = exchange_name

    async def publish(self, message: BaseModel, topic: str) -> None:
        exchange = await self.channel.declare_exchange(
            self.exchange_name,
            type=aio_pika.ExchangeType.TOPIC,
        )
        routing_key = f"{self.exchange_name}.{topic}."
        await exchange.publish(
            aio_pika.Message(body=message.model_dump_json().encode()),
            routing_key=routing_key,
        )


class AIOPikaConnectionFactory(MessageBusConnectionFactory):

    def __init__(
        self,
        url: str,
        exchange: str,
    ):
        self.url = url
        self.exchange = exchange

    @asynccontextmanager
    async def provide_connection(self) -> AsyncGenerator[AIOPikaMessagePublisher, Any]:
        connection = await aio_pika.connect_robust(self.url)
        async with connection:
            async with connection.channel(publisher_confirms=False) as channel:

                tx = channel.transaction()

                await tx.select()

                try:
                    yield AIOPikaMessagePublisher(channel, exchange_name=self.exchange)
                    await tx.commit()
                except Exception as e:
                    await tx.rollback()
                    raise e
